@@ -25,7 +25,7 @@ def at_date(
     """
     if get_bounds:
         return timedelta.min, timedelta.max
-    return event_start - knowledge_time
+    return event_start - knowledge_time.astimezone(event_start.tzinfo)
 
 
 def ex_post(
@@ -149,3 +149,7 @@ def determine_ex_ante_knowledge_horizon_for_x_days_ago_at_y_oclock(
         FutureWarning,
     )
     return x_days_ago_at_y_oclock(event_start, *args, **kwargs)
+
+
+# aliases
+x_days_ago_at_y_o_clock = x_days_ago_at_y_oclock  # forgive a common typo
