@@ -4,10 +4,10 @@ from typing import Optional, Tuple, Union
 
 from isodate import (
     ISO8601Error,
+    datetime_isoformat,
     duration_isoformat,
     parse_datetime,
     parse_duration,
-    time_isoformat,
 )
 from pytz import timezone
 
@@ -19,8 +19,8 @@ def datetime_x_days_ago_at_y_oclock(
 ) -> datetime:
     """Returns the datetime x days ago a y o'clock as determined from the perspective of timezone z."""
     if isinstance(y, float):
-        micros = int(y * 60 * 60 * 10 ** 6)
-        s, micros = divmod(micros, 10 ** 6)
+        micros = int(y * 60 * 60 * 10**6)
+        s, micros = divmod(micros, 10**6)
         m, s = divmod(s, 60)
         h, m = divmod(m, 60)
     else:
@@ -43,7 +43,7 @@ def jsonify_time_dict(d: dict) -> dict:
     d2 = {}
     for k, v in d.items():
         if isinstance(v, datetime):
-            d2[k] = time_isoformat(v)
+            d2[k] = datetime_isoformat(v)
         elif isinstance(v, timedelta):
             d2[k] = duration_isoformat(v)
         else:
